import tempfile
import numpy as np
import pytest
import pyccl as ccl
# import warnings  # TODO: Uncomment for CCLv3.


def test_parameters_lcdmDefaultParams():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96)

    assert np.allclose(cosmo['Omega_c'], 0.25)
    assert np.allclose(cosmo['Omega_b'], 0.05)
    assert np.allclose(cosmo['Omega_m'], 0.30)
    assert np.allclose(cosmo['Omega_k'], 0)
    assert np.allclose(cosmo['sqrtk'], 0)
    assert np.allclose(cosmo['k_sign'], 0)
    assert np.allclose(cosmo['w0'], -1)
    assert np.allclose(cosmo['wa'], 0)
    assert np.allclose(cosmo['H0'], 70)
    assert np.allclose(cosmo['h'], 0.7)
    assert np.allclose(cosmo['A_s'], 2.1e-9)
    assert np.allclose(cosmo['n_s'], 0.96)
    assert np.isnan(cosmo['sigma8'])
    assert np.allclose(cosmo['Neff'], 3.044)
    assert cosmo['N_nu_mass'] == 0
    assert np.allclose(cosmo['N_nu_rel'], 3.044)
    assert np.allclose(cosmo['sum_nu_masses'], 0)
    assert np.allclose(cosmo['m_nu'], 0)
    assert np.allclose(cosmo['Omega_nu_mass'], 0)
    assert np.allclose(cosmo['T_CMB'], ccl.cosmology.DefaultParams.T_CMB)

    # these are defined in the code via some constants so
    # going to test the total
    #     Omega_nu_rel
    #     Omega_g
    #     Omega_l
    assert np.allclose(
        cosmo['Omega_l'] + cosmo['Omega_m'] + cosmo['Omega_g'] +
        cosmo['Omega_nu_rel'] + cosmo['Omega_nu_mass'] + cosmo['Omega_k'],
        1)


@pytest.mark.parametrize('mass_split', ['normal', 'inverted', 'single'])
def test_parameters_nu(mass_split):
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        wa=0.01,
        w0=-1,
        Neff=3.044,
        Omega_k=0.0,
        m_nu=0.15,
        mass_split=mass_split
    )

    if mass_split == 'inverted':
        assert np.allclose(cosmo['m_nu'][1]**2 - cosmo['m_nu'][0]**2,
                           ccl.physical_constants.DELTAM12_sq,
                           atol=1e-4, rtol=0)
        assert np.allclose(
            cosmo['m_nu'][2]**2 - cosmo['m_nu'][0]**2,
            ccl.physical_constants.DELTAM13_sq_neg, atol=1e-4, rtol=0)
    elif mass_split == 'normal':
        assert np.allclose(cosmo['m_nu'][1]**2 - cosmo['m_nu'][0]**2,
                           ccl.physical_constants.DELTAM12_sq,
                           atol=1e-4, rtol=0)
        assert np.allclose(
            cosmo['m_nu'][2]**2 - cosmo['m_nu'][0]**2,
            ccl.physical_constants.DELTAM13_sq_pos, atol=1e-4, rtol=0)
    elif mass_split == 'single':
        assert len(cosmo["m_nu"]) == 1
        assert np.allclose(cosmo['m_nu'][0], 0.15, atol=1e-4, rtol=0)


def test_parameters_nu_Nnurel_neg():
    with pytest.raises(ValueError):
        ccl.Cosmology(
            Omega_c=0.27, Omega_b=0.049,
            h=0.67, sigma8=0.8, n_s=0.96, m_nu=[0.03, 0.02, 0.04],
            Neff=3., mass_split='list')


def test_parameters_nu_list():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        m_nu=[0.1, 0.01, 0.003],
        mass_split='list')

    assert np.allclose(cosmo['Omega_c'], 0.25)
    assert np.allclose(cosmo['Omega_b'], 0.05)
    assert np.allclose(cosmo['Omega_m'] - cosmo['Omega_nu_mass'], 0.30)
    assert np.allclose(cosmo['Omega_k'], 0)
    assert np.allclose(cosmo['sqrtk'], 0)
    assert np.allclose(cosmo['k_sign'], 0)
    assert np.allclose(cosmo['w0'], -1)
    assert np.allclose(cosmo['wa'], 0)
    assert np.allclose(cosmo['H0'], 70)
    assert np.allclose(cosmo['h'], 0.7)
    assert np.allclose(cosmo['A_s'], 2.1e-9)
    assert np.allclose(cosmo['n_s'], 0.96)
    assert np.isnan(cosmo['sigma8'])
    assert np.allclose(cosmo['T_CMB'], ccl.cosmology.DefaultParams.T_CMB)

    # these are defined in the code via some constants so
    # going to test the total
    #     Omega_nu_rel
    #     Omega_g
    #     Omega_l
    assert np.allclose(
        cosmo['Omega_l'] + cosmo['Omega_m'] + cosmo['Omega_g'] +
        cosmo['Omega_nu_rel'] + cosmo['Omega_k'],
        1)

    assert np.allclose(cosmo['Neff'], 3.044)
    assert cosmo['N_nu_mass'] == 3
    assert np.allclose(cosmo['sum_nu_masses'], 0.1 + 0.01 + 0.003)
    assert np.allclose(cosmo['m_nu'], [0.1, 0.01, 0.003])


def test_parameters_nu_normal():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        m_nu=0.3,
        mass_split='normal')

    assert np.allclose(cosmo['Omega_c'], 0.25)
    assert np.allclose(cosmo['Omega_b'], 0.05)
    assert np.allclose(cosmo['Omega_m'] - cosmo['Omega_nu_mass'], 0.30)
    assert np.allclose(cosmo['Omega_k'], 0)
    assert np.allclose(cosmo['sqrtk'], 0)
    assert np.allclose(cosmo['k_sign'], 0)
    assert np.allclose(cosmo['w0'], -1)
    assert np.allclose(cosmo['wa'], 0)
    assert np.allclose(cosmo['H0'], 70)
    assert np.allclose(cosmo['h'], 0.7)
    assert np.allclose(cosmo['A_s'], 2.1e-9)
    assert np.allclose(cosmo['n_s'], 0.96)
    assert np.isnan(cosmo['sigma8'])
    assert np.allclose(cosmo['T_CMB'], ccl.cosmology.DefaultParams.T_CMB)

    # these are defined in the code via some constants so
    # going to test the total
    #     Omega_nu_rel
    #     Omega_g
    #     Omega_l
    assert np.allclose(
        cosmo['Omega_l'] + cosmo['Omega_m'] + cosmo['Omega_g'] +
        cosmo['Omega_nu_rel'] + cosmo['Omega_k'],
        1)

    assert np.allclose(cosmo['Neff'], 3.044)
    assert cosmo['N_nu_mass'] == 3
    assert np.allclose(cosmo['sum_nu_masses'], 0.3)


def test_parameters_nu_inverted():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        m_nu=0.3,
        mass_split='inverted')

    assert np.allclose(cosmo['Omega_c'], 0.25)
    assert np.allclose(cosmo['Omega_b'], 0.05)
    assert np.allclose(cosmo['Omega_m'] - cosmo['Omega_nu_mass'], 0.30)
    assert np.allclose(cosmo['Omega_k'], 0)
    assert np.allclose(cosmo['sqrtk'], 0)
    assert np.allclose(cosmo['k_sign'], 0)
    assert np.allclose(cosmo['w0'], -1)
    assert np.allclose(cosmo['wa'], 0)
    assert np.allclose(cosmo['H0'], 70)
    assert np.allclose(cosmo['h'], 0.7)
    assert np.allclose(cosmo['A_s'], 2.1e-9)
    assert np.allclose(cosmo['n_s'], 0.96)
    assert np.isnan(cosmo['sigma8'])
    assert np.allclose(cosmo['T_CMB'], ccl.cosmology.DefaultParams.T_CMB)

    # these are defined in the code via some constants so
    # going to test the total
    #     Omega_nu_rel
    #     Omega_g
    #     Omega_l
    assert np.allclose(
        cosmo['Omega_l'] + cosmo['Omega_m'] + cosmo['Omega_g'] +
        cosmo['Omega_nu_rel'] + cosmo['Omega_k'],
        1)

    assert np.allclose(cosmo['Neff'], 3.044)
    assert cosmo['N_nu_mass'] == 3
    assert np.allclose(cosmo['sum_nu_masses'], 0.3)


def test_parameters_nu_equal():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        m_nu=0.3,
        mass_split='equal')

    assert np.allclose(cosmo['Omega_c'], 0.25)
    assert np.allclose(cosmo['Omega_b'], 0.05)
    assert np.allclose(cosmo['Omega_m'] - cosmo['Omega_nu_mass'], 0.30)
    assert np.allclose(cosmo['Omega_k'], 0)
    assert np.allclose(cosmo['sqrtk'], 0)
    assert np.allclose(cosmo['k_sign'], 0)
    assert np.allclose(cosmo['w0'], -1)
    assert np.allclose(cosmo['wa'], 0)
    assert np.allclose(cosmo['H0'], 70)
    assert np.allclose(cosmo['h'], 0.7)
    assert np.allclose(cosmo['A_s'], 2.1e-9)
    assert np.allclose(cosmo['n_s'], 0.96)
    assert np.isnan(cosmo['sigma8'])
    assert np.allclose(cosmo['T_CMB'], ccl.cosmology.DefaultParams.T_CMB)

    # these are defined in the code via some constants so
    # going to test the total
    #     Omega_nu_rel
    #     Omega_g
    #     Omega_l
    assert np.allclose(
        cosmo['Omega_l'] + cosmo['Omega_m'] + cosmo['Omega_g'] +
        cosmo['Omega_nu_rel'] + cosmo['Omega_k'],
        1)

    assert np.allclose(cosmo['Neff'], 3.044)
    assert cosmo['N_nu_mass'] == 3
    assert np.allclose(cosmo['sum_nu_masses'], 0.3)
    assert np.allclose(cosmo['m_nu'], 0.1)


@pytest.mark.parametrize('m_nu,kind', [(0.05, 'normal'), (0.09, 'inverted')])
def test_parameters_nu_unphysical_raises(m_nu, kind):
    with pytest.raises(ValueError):
        ccl.Cosmology(
            Omega_c=0.25,
            Omega_b=0.05,
            h=0.7,
            A_s=2.1e-9,
            n_s=0.96,
            m_nu=m_nu,
            mass_split=kind)


def test_parameters_missing():
    """
    Check that errors are raised when compulsory parameters are missing, but
    not when non-compulsory ones are.
    """

    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25)

    # Check that a single missing compulsory parameter is noticed
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9)
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.96)
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, A_s=2.1e-9, n_s=0.96)
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, h=0.7, A_s=2.1e-9, n_s=0.96)
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96)

    # Check that sigma8 vs A_s is handled ok.
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.8,
                      A_s=2.1e-9, sigma8=0.7)
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.8)


def test_parameters_read_write():
    """Check that Cosmology objects can be read and written"""
    params = ccl.Cosmology(
        Omega_c=0.25, Omega_b=0.05, h=0.7, A_s=2.1e-9, n_s=0.96,
        m_nu=[0.02, 0.1, 0.05], mass_split='list')

    # Make a temporary file name
    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
        temp_file_name = tmpfile.name

    # Write out and then read in the parameters from that file
    with pytest.raises(IOError):
        params.write_yaml("/bogus/file/name")
    params.write_yaml(temp_file_name)
    params2 = ccl.Cosmology.read_yaml(temp_file_name)

    # Check the read-in params are equal to the written out ones
    assert params["Omega_c"] == params2["Omega_c"]
    assert params["Neff"] == params2["Neff"]
    assert params["sum_nu_masses"] == params2["sum_nu_masses"]

    # check overriding parameters with kwargs
    params3 = ccl.Cosmology.read_yaml(temp_file_name,
                                      matter_power_spectrum='linear',
                                      n_s=1.1)
    # check unmodified parameters are the same
    assert params["Omega_c"] == params3["Omega_c"]
    assert params["Neff"] == params3["Neff"]
    # check new parameters and config correctly updated
    assert params3["n_s"] == 1.1
    assert params["sum_nu_masses"] == params3["sum_nu_masses"]
    assert params3._config_init_kwargs['matter_power_spectrum'] == 'linear'

    # Now make a file that will be deleted so it does not exist
    # and check the right error is raise
    with tempfile.NamedTemporaryFile(delete=True) as tmpfile:
        temp_file_name = tmpfile.name

    with pytest.raises(IOError):
        ccl.Cosmology.read_yaml(filename=temp_file_name)
    with pytest.raises(IOError):
        params.read_yaml(
            filename=temp_file_name+"/nonexistent_directory/params.yml")


def test_omega_k():
    """ Check that the value of Omega_k is within reasonable bounds. """
    with pytest.raises(ValueError):
        ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7,
                      A_s=2.1e-9, n_s=0.96, Omega_k=-2)



import numpy as np
import pytest
import pyccl as ccl
from pyccl import CCLWarning
from .test_cclobject import check_eq_repr_hash

COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
COSMO.compute_distances()  # needed to suppress C-level warnings


def test_Tracer_eq_repr_hash():
    # Test eq, repr, hash for Tracer.
    # empty Tracer
    assert check_eq_repr_hash(ccl.Tracer(), ccl.Tracer())

    # no transfer
    TR1 = ccl.CMBLensingTracer(COSMO, z_source=1101)
    TR2 = ccl.CMBLensingTracer(COSMO, z_source=1101)
    assert check_eq_repr_hash(TR1, TR2)

    TR3 = ccl.CMBLensingTracer(COSMO, z_source=1100)
    assert check_eq_repr_hash(TR1, TR3, equal=False)

    # transfer_fka
    lk = np.linspace(-5, 1, 32)
    a = np.linspace(0.5, 1, 4)
    tka = np.ones((a.size, lk.size))
    TR1.add_tracer(COSMO, transfer_ka=(a, lk, tka))
    TR2.add_tracer(COSMO, transfer_ka=(a, lk, 2*tka))
    assert check_eq_repr_hash(TR1, TR2, equal=False)

    # transfer_fk
    TR4 = ccl.CMBLensingTracer(COSMO, z_source=1101)
    TR5 = ccl.CMBLensingTracer(COSMO, z_source=1101)
    TR4.add_tracer(COSMO, transfer_k=(lk, np.ones_like(lk)))
    TR5.add_tracer(COSMO, transfer_k=(lk, 2*np.ones_like(lk)))
    assert check_eq_repr_hash(TR4, TR5, equal=False)

    # edge-case: different type
    assert check_eq_repr_hash(ccl.Tracer(), 1, equal=False)

    # edge-case: different `der_angles` & `der_bessel`
    t1, t2 = ccl.Tracer(), ccl.Tracer()
    t1.add_tracer(COSMO, der_angles=0, der_bessel=0)
    t2.add_tracer(COSMO, der_angles=1, der_bessel=1)
    assert check_eq_repr_hash(t1, t2, equal=False)

    # edge-case: only one has kernel
    chi = np.linspace(0, 50, 16)
    t1, t2 = ccl.Tracer(), ccl.Tracer()
    t1.add_tracer(COSMO, kernel=(chi, np.ones_like(chi)))
    t2.add_tracer(COSMO)
    assert check_eq_repr_hash(t1, t2, equal=False)

    # edge-case: only one has transfer
    t1, t2 = ccl.Tracer(), ccl.Tracer()
    t1.add_tracer(COSMO, transfer_k=(lk, np.ones_like(lk)))
    t2.add_tracer(COSMO)
    assert check_eq_repr_hash(t1, t2, equal=False)

    # edge-case: only one has specific transfer type
    t1, t2 = ccl.Tracer(), ccl.Tracer()
    t1.add_tracer(COSMO, transfer_k=(lk, np.ones_like(lk)))
    t2.add_tracer(COSMO, transfer_a=(a, np.ones_like(a)))
    assert check_eq_repr_hash(t1, t2, equal=False)


def dndz(z):
    return np.exp(-((z-0.5)/0.1)**2)


def get_tracer(tracer_type, cosmo=None, **tracer_kwargs):
    if cosmo is None:
        cosmo = COSMO
    z = np.linspace(0., 1., 2000)
    n = dndz(z)
    b = np.sqrt(1. + z)

    if tracer_type == 'nc':
        ntr = 3
        tr = ccl.NumberCountsTracer(cosmo, has_rsd=True,
                                    dndz=(z, n),
                                    bias=(z, b),
                                    mag_bias=(z, b),
                                    **tracer_kwargs)
    elif tracer_type == 'wl':
        ntr = 2
        tr = ccl.WeakLensingTracer(cosmo,
                                   dndz=(z, n),
                                   ia_bias=(z, b),
                                   **tracer_kwargs)
    elif tracer_type == 'cl':
        ntr = 1
        tr = ccl.CMBLensingTracer(cosmo, z_source=1100., **tracer_kwargs)
    else:
        ntr = 0
        tr = ccl.Tracer(**tracer_kwargs)
    return tr, ntr


def test_tracer_mag_0p4():
    z = np.linspace(0., 1., 2000)
    n = dndz(z)
    # Small bias so the magnification contribution
    # is significant (if there at all)
    b = np.ones_like(z)*0.1
    s_no = np.ones_like(z)*0.4
    s_yes = np.zeros_like(z)
    # Tracer with no magnification by construction
    t1 = ccl.NumberCountsTracer(COSMO, has_rsd=True,
                                dndz=(z, n),
                                bias=(z, b))
    # Tracer with s=0.4
    t2 = ccl.NumberCountsTracer(COSMO, has_rsd=True,
                                dndz=(z, n),
                                bias=(z, b),
                                mag_bias=(z, s_no))
    # Tracer with magnification
    t3 = ccl.NumberCountsTracer(COSMO, has_rsd=True,
                                dndz=(z, n),
                                bias=(z, b),
                                mag_bias=(z, s_yes))
    ls = np.array([2, 200, 2000])
    cl1 = ccl.angular_cl(COSMO, t1, t1, ls)
    cl2 = ccl.angular_cl(COSMO, t2, t2, ls)
    cl3 = ccl.angular_cl(COSMO, t3, t3, ls)
    # Check cl1 == cl2
    assert np.all(np.fabs(cl2/cl1-1) < 1E-5)
    # Check cl1 != cl3
    assert np.all(np.fabs(cl3/cl1-1) > 1E-2)


@pytest.mark.parametrize('tracer_type', ['nc', 'wl'])
def test_tracer_dndz_smoke(tracer_type):
    tr, _ = get_tracer(tracer_type)
    for z in [np.linspace(0.5, 0.6, 10),
              0.5]:
        n1 = dndz(z)
        n2 = tr.get_dndz(z)
        assert np.all(np.fabs(n1 / n2 - 1) < 1E-5)


@pytest.mark.parametrize('tracer_type', ['nc', 'wl', 'cl', 'not'])
def test_tracer_kernel_smoke(tracer_type):
    tr, ntr = get_tracer(tracer_type)
    for chi in [np.linspace(0., 3000., 128),
                [100., 1000.],
                100.]:
        w = tr.get_kernel(chi)

        assert w.shape[0] == ntr
        if ntr > 0:
            for ww in w:
                assert np.shape(ww) == np.shape(chi)


@pytest.mark.parametrize('tracer_type', ['nc', 'wl', 'cl', 'not'])
def test_tracer_der_bessel_smoke(tracer_type):
    tr, ntr = get_tracer(tracer_type)

    if tracer_type == 'nc':
        dd = np.array([0, 2, -1])
    elif tracer_type == 'wl':
        dd = np.array([-1, -1])
    elif tracer_type == 'cl':
        dd = np.array([-1])
    else:
        dd = np.array([])

    d = tr.get_bessel_derivative()
    assert np.all(d == dd)


@pytest.mark.parametrize('tracer_type', ['nc', 'wl', 'cl', 'not'])
def test_tracer_f_ell_smoke(tracer_type):
    tr, ntr = get_tracer(tracer_type)
    for ell in [np.linspace(0., 3000., 128),
                [100., 1000.],
                100.]:
        fl = tr.get_f_ell(ell)

        assert fl.shape[0] == ntr
        if ntr > 0:
            for f in fl:
                assert np.shape(f) == np.shape(ell)


@pytest.mark.parametrize('tracer_type', ['nc', 'wl', 'cl', 'not'])
def test_tracer_transfer_smoke(tracer_type):
    tr, ntr = get_tracer(tracer_type)
    for lk in [np.linspace(-3., 1., 10),
               [-2., 0.],
               -1.]:
        for a in [np.linspace(0.5, 1., 8),
                  [0.4, 1.],
                  0.9]:
            tf = tr.get_transfer(lk, a)

            assert tf.shape[0] == ntr
            if ntr > 0:
                if np.ndim(a) == 0:
                    if np.ndim(lk) == 0:
                        shap = (ntr, )
                    else:
                        shap = (ntr, len(lk))
                else:
                    if np.ndim(lk) == 0:
                        shap = (ntr, len(a))
                    else:
                        shap = (ntr, len(lk), len(a))
            else:
                shap = (0, )
            assert tf.shape == shap


def test_tracer_nz_support():
    z_max = 1.0
    a = np.linspace(1/(1+z_max), 1.0, 100)

    background_def = {"a": a,
                      "chi": ccl.comoving_radial_distance(COSMO, a),
                      "h_over_h0": ccl.h_over_h0(COSMO, a)}

    calculator_cosmo = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.045, h=0.67,
        sigma8=0.8, n_s=0.96,
        background=background_def)

    z = np.linspace(0., 2., 2000)
    n = dndz(z)

    with pytest.raises(ValueError):
        _ = ccl.WeakLensingTracer(calculator_cosmo, dndz=(z, n))

    with pytest.raises(ValueError):
        _ = ccl.NumberCountsTracer(calculator_cosmo, has_rsd=False,
                                   dndz=(z, n), bias=(z, np.ones_like(z)))

    with pytest.raises(ValueError):
        _ = ccl.CMBLensingTracer(calculator_cosmo, z_source=2.0)


def new_simple_cosmo():
    return ccl.CosmologyVanillaLCDM(transfer_function='bbks',
                                    matter_power_spectrum="linear")


def test_tracer_nz_norm_spline_vs_gsl_intergation():
    # Create a new Cosmology object so that we're not messing with the other
    # tests
    ccl.gsl_params.NZ_NORM_SPLINE_INTEGRATION = True
    cosmo = new_simple_cosmo()
    tr_wl, _ = get_tracer("wl", cosmo)
    tr_nc, _ = get_tracer("nc", cosmo)

    w_wl_spline, _ = tr_wl.get_kernel(chi=None)
    w_nc_spline, _ = tr_nc.get_kernel(chi=None)

    ccl.gsl_params.NZ_NORM_SPLINE_INTEGRATION = False
    cosmo = new_simple_cosmo()
    tr_wl, _ = get_tracer("wl", cosmo)
    tr_nc, _ = get_tracer("nc", cosmo)

    w_wl_gsl, _ = tr_wl.get_kernel(chi=None)
    w_nc_gsl, _ = tr_nc.get_kernel(chi=None)

    for w_spline, w_gsl in zip(w_wl_spline, w_wl_gsl):
        assert np.allclose(w_spline, w_gsl, atol=0, rtol=1e-8)
    for w_spline, w_gsl in zip(w_nc_spline, w_nc_gsl):
        assert np.allclose(w_spline, w_gsl, atol=0, rtol=1e-8)

    ccl.gsl_params.reload()  # reset to the default parameters


@pytest.mark.parametrize('z_min, z_max, n_z_samples', [(0.0, 1.0, 2000),
                                                       (0.0, 1.0, 1000),
                                                       (0.0, 1.0, 500),
                                                       (0.0, 1.0, 256),
                                                       (0.3, 1.0, 1000)])
def test_tracer_lensing_kernel_spline_vs_gsl_intergation(z_min, z_max,
                                                         n_z_samples):
    # Create a new Cosmology object so that we're not messing with the other
    # tests
    z = np.linspace(z_min, z_max, n_z_samples)
    n = dndz(z)

    # Make sure case where z[0] > 0 and n[0] > 0 is tested for
    if z_min > 0:
        assert n[0] > 0

    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = True
    cosmo = new_simple_cosmo()
    tr_wl = ccl.WeakLensingTracer(cosmo, dndz=(z, n))
    w_wl_spline, _ = tr_wl.get_kernel(chi=None)

    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
    cosmo = new_simple_cosmo()
    tr_wl = ccl.WeakLensingTracer(cosmo, dndz=(z, n))
    w_wl_gsl, chi = tr_wl.get_kernel(chi=None)

    # Peak of kernel is ~1e-5
    if n_z_samples >= 1000:
        assert np.allclose(w_wl_spline[0], w_wl_gsl[0], atol=1e-10, rtol=1e-9)
    else:
        assert np.allclose(w_wl_spline[0], w_wl_gsl[0], atol=5e-9, rtol=1e-5)

    ccl.gsl_params.reload()  # reset to the default parameters


@pytest.mark.parametrize('z_min, z_max, n_z_samples', [(0.0, 1.0, 2000),
                                                       (0.0, 1.0, 1000),
                                                       (0.0, 1.0, 500),
                                                       (0.0, 1.0, 256),
                                                       (0.3, 1.0, 1000)])
def test_tracer_magnification_kernel_spline_vs_gsl_intergation(z_min, z_max,
                                                               n_z_samples):
    # Create a new Cosmology object so that we're not messing with the other
    # tests
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.045, h=0.67,
                          sigma8=0.8, n_s=0.96,
                          transfer_function='bbks',
                          matter_power_spectrum='linear')
    z = np.linspace(z_min, z_max, n_z_samples)
    n = dndz(z)
    b = np.zeros_like(z)

    # Make sure case where z[0] > 0 and n[0] > 0 is tested for
    if z_min > 0:
        assert n[0] > 0

    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = True
    tr_mg = ccl.NumberCountsTracer(cosmo, has_rsd=False, dndz=(z, n),
                                   bias=(z, b), mag_bias=(z, b))
    w_mg_spline, _ = tr_mg.get_kernel(chi=None)
    ccl.gsl_params.reload()

    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = True
    tr_mg = ccl.NumberCountsTracer(cosmo, has_rsd=False, dndz=(z, n),
                                   bias=(z, b), mag_bias=(z, b))
    w_mg_gsl, chi = tr_mg.get_kernel(chi=None)
    tr_wl = ccl.WeakLensingTracer(cosmo, dndz=(z, n))
    w_wl_gsl, _ = tr_wl.get_kernel(chi=None)
    ccl.gsl_params.reload()

    # Peak of kernel is ~1e-5
    if n_z_samples >= 1000:
        assert np.allclose(w_mg_spline[1], w_mg_gsl[1],
                           atol=2e-10, rtol=1e-9)
        assert np.allclose(w_mg_spline[1], -2*w_wl_gsl[0],
                           atol=2e-10, rtol=1e-9)
    else:
        assert np.allclose(w_mg_spline[1], w_mg_gsl[1],
                           atol=1e-8, rtol=1e-5)
        assert np.allclose(w_mg_spline[1], -2*w_wl_gsl[0],
                           atol=1e-8, rtol=1e-5)


def test_tracer_delta_function_nz():
    z = np.linspace(0., 1., 2000)
    z_s_idx = int(z.size*0.8)
    z_s = z[z_s_idx]
    n = np.zeros_like(z)
    n[z_s_idx] = 2.0

    tr_wl = ccl.WeakLensingTracer(COSMO, dndz=(z, n))

    # Single source plane tracer to compare against
    chi_kappa, w_kappa = ccl.tracers.get_kappa_kernel(COSMO, z_source=z_s,
                                                      n_samples=100)

    # Use the same comoving distances
    w = tr_wl.get_kernel(chi=chi_kappa)

    assert np.allclose(w[0], w_kappa, atol=1e-8, rtol=1e-6)
    # at z=z_source, interpolation becomes apparent, so for this test we
    # ignore these data points.
    assert np.allclose(w[0][:-2], w_kappa[:-2], atol=1e-11, rtol=1e-11)


@pytest.mark.parametrize('tracer_type', ['nc', 'wl', 'cl'])
def test_tracer_n_sample_smoke(tracer_type):
    tr, ntr = get_tracer(tracer_type, n_samples=50)
    if tracer_type != "cl":
        # n_samples=None should fall back to using the samples from the n(z).
        tr, ntr = get_tracer(tracer_type, n_samples=None)


def test_tracer_n_sample_wl():
    z = np.linspace(0., 1., 2000)
    n = dndz(z)

    n_samples = 50
    tr_wl = ccl.WeakLensingTracer(COSMO, dndz=(z, n), n_samples=n_samples)
    w, chi = tr_wl.get_kernel(chi=None)

    assert w[0].shape[-1] == n_samples
    assert chi[0].shape[-1] == n_samples


def test_tracer_n_sample_warn():
    z = np.linspace(0., 1., 50)
    n = dndz(z)

    with pytest.warns(CCLWarning):
        _ = ccl.WeakLensingTracer(COSMO, dndz=(z, n))


def test_tracer_bool():
    assert bool(ccl.Tracer()) is False
    assert bool(ccl.CMBLensingTracer(COSMO, z_source=1100)) is True


def test_tracer_chi_min_max():
    # Test that it can access the C-level chi_min and chi_max.
    tr = ccl.CMBLensingTracer(COSMO, z_source=1100)
    assert tr.chi_min == tr._trc[0].chi_min
    assert tr.chi_max == tr._trc[0].chi_max

    # Returns the lowest/highest if chi_min or chi_max are not the same.
    chi = np.linspace(tr.chi_min+0.05, tr.chi_max+0.05, 128)
    wchi = np.ones_like(chi)
    tr.add_tracer(COSMO, kernel=(chi, wchi))
    assert tr.chi_min == tr._trc[0].chi_min
    assert tr.chi_max == tr._trc[1].chi_max


def test_zpower_raises():
    with pytest.raises(ValueError):
        ccl.Tracer.from_z_power(COSMO, A=1., z_min=1.0,
                                z_max=0.1, alpha=1.0)


def test_tracer_increase_sf():
    z = np.linspace(0, 3., 32)
    one = np.ones(len(z))
    chi = ccl.comoving_radial_distance(COSMO, 1./(1+z))
    sf = 1./(1+z)
    tr = ccl.Tracer()
    with pytest.raises(ValueError):
        tr.add_tracer(COSMO, kernel=(chi, one),
                      transfer_a=(sf, one))
    # Check it works in the right order
    tr.add_tracer(COSMO, kernel=(chi, one),
                  transfer_a=(sf[::-1], one))


def test_tracer_repr():
    """Check that the repr works as intended."""
    # Equal Tracers with a kernel.
    tr1 = ccl.CMBLensingTracer(COSMO, z_source=1100)
    tr2 = ccl.CMBLensingTracer(COSMO, z_source=1100)
    assert tr1 == tr2
    # Unequal.
    tr3 = ccl.CMBLensingTracer(COSMO, z_source=1101)
    assert tr3 != tr1
    # We add an extra tracer and check unequal.
    chi = np.linspace(tr1.chi_min, tr1.chi_max, 128)
    wchi = np.ones_like(chi)
    tr2.add_tracer(COSMO, kernel=(chi, wchi))
    assert tr2 != tr1
    # Check empty tracer.
    z = np.linspace(0, 0.5, 128)
    nz = np.ones_like(z)
    tr4 = ccl.NzTracer()
    tr5 = ccl.NumberCountsTracer(COSMO, dndz=(z, nz), has_rsd=False)  # all off
    assert tr4 == tr5
    # Check tracers with transfer functions.
    # transfer_a
    tr6 = ccl.NumberCountsTracer(COSMO, dndz=(z, nz), has_rsd=True)
    tr7 = ccl.NumberCountsTracer(COSMO, dndz=(z, nz), has_rsd=True)
    assert tr6 == tr7
    # transfer_k
    lk = np.linspace(-5, 2, 8)
    t_k = np.ones_like(lk)
    tr6.add_tracer(COSMO, transfer_k=(lk, t_k))
    tr7.add_tracer(COSMO, transfer_k=(lk, t_k))
    assert tr6 == tr7
    # transfer_a
    tr6.add_tracer(COSMO, transfer_a=(1/(1+z)[::-1], nz))
    tr7.add_tracer(COSMO, transfer_a=(1/(1+z)[::-1], nz))
    assert tr6 == tr7
    # transfer_ka
    a = np.linspace(0.5, 1.0, 8)
    t_ka = np.ones((a.size, lk.size))
    tr6.add_tracer(COSMO, transfer_ka=(a, lk, t_ka))
    tr7.add_tracer(COSMO, transfer_ka=(a, lk, t_ka))
    assert tr6 == tr7
    # different extrap orders
    tr6.add_tracer(COSMO, transfer_ka=(a, lk, t_ka), extrap_order_lok=0)
    tr7.add_tracer(COSMO, transfer_ka=(a, lk, t_ka), extrap_order_lok=1)
    assert tr6 != tr7



import numpy as np
import pyccl as ccl
import pytest

from pyccl.errors import CCLError

COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='halofit')

COSMO_LOWS8 = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.1, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='halofit')


@pytest.mark.parametrize("cosmo", [COSMO, COSMO_LOWS8])
def test_halofit_highz(cosmo):
    vals = [(25, 75)] + list(zip(range(0, 98), range(1, 99)))
    for zl, zh in vals:
        al = 1.0/(1 + zl)
        ah = 1.0/(1 + zh)

        k = np.logspace(0, 2, 10)
        pkratl = (
            ccl.nonlin_matter_power(cosmo, k, al)
            / ccl.linear_matter_power(cosmo, k, al)
        )
        pkrath = (
            ccl.nonlin_matter_power(cosmo, k, ah)
            / ccl.linear_matter_power(cosmo, k, ah)
        )

        assert np.all(pkratl >= pkrath), (zl, zh, pkratl, pkrath)


def test_halofit_background_check():
    ccl.spline_params.A_SPLINE_MIN = 0.4
    ccl.spline_params.A_SPLINE_MINLOG = 0.3
    ccl.spline_params.A_SPLINE_MIN_PK = 0.4
    ccl.spline_params.A_SPLINE_MINLOG_PK = 0.3

    cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7,
                          n_s=0.97,
                          sigma8=0.8,
                          w0=-1.04, wa=-0.1,
                          matter_power_spectrum="halofit",
                          transfer_function="eisenstein_hu")

    k = np.geomspace(1e-3, 1, 10)

    with pytest.raises(CCLError):
        ccl.nonlin_matter_power(cosmo, k, a=0.5)

    ccl.spline_params.reload()



__all__ = ("HaloBiasTinker10",)

import numpy as np

from . import HaloBias, get_delta_c


class HaloBiasTinker10(HaloBias):
    """Implements halo bias as described in `Tinker et al. 2010
    <https://arxiv.org/abs/1001.3162>`_. This parametrization accepts S.O.
    masses with :math:`200 < \\Delta < 3200`, defined with respect to the
    matter density. This can be automatically translated to S.O. masses
    defined with respect to the critical density.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = "Tinker10"

    def __init__(self, *,
                 mass_def="200m",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        # True for FoF since Tinker10 is not defined for this mass def.
        return mass_def.Delta == "fof"

    def _setup(self):
        self.B = 0.183
        self.b = 1.5
        self.c = 2.4
        self.dc = get_delta_c(None, None, kind='EdS')

    def _get_bsigma(self, cosmo, sigM, a):
        nu = self.dc / sigM
        ld = np.log10(self.mass_def._get_Delta_m(cosmo, a))
        xp = np.exp(-(4./ld)**4.)
        A = 1.0 + 0.24 * ld * xp
        C = 0.019 + 0.107 * ld + 0.19*xp
        aa = 0.44 * ld - 0.88
        nupa = nu**aa
        return 1 - A * nupa / (nupa + self.dc**aa) + (
            self.B * nu**self.b + C * nu**self.c)



from .halo_model_base import *
from .concentration import *
from .hbias import *
from .hmfunc import *
from .massdef import *
from .profiles import *
from .profiles_2pt import *
from .pk_1pt import *
from .pk_2pt import *
from .pk_4pt import *
from .halo_model import *



import numpy as np
import pytest
import pyccl as ccl


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='halofit')


@pytest.mark.parametrize('k', [
    1,
    1.0,
    [0.3, 0.5, 10],
    np.array([0.3, 0.5, 10])])
def test_linear_matter_power_smoke(k):
    a = 0.8
    pk = ccl.linear_matter_power(COSMO, k, a)
    assert np.all(np.isfinite(pk))
    assert np.shape(pk) == np.shape(k)


def test_linear_power_raises():
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function='bbks')
    with pytest.raises(KeyError):
        ccl.linear_power(cosmo, 1., 1., p_of_k_a='a:b')


def test_nonlin_power_raises():
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function='bbks')
    with pytest.raises(KeyError):
        ccl.nonlin_power(cosmo, 1., 1., p_of_k_a='a:b')


@pytest.mark.parametrize('k', [
    1,
    1.0,
    [0.3, 0.5, 10],
    np.array([0.3, 0.5, 10])])
def test_nonlin_matter_power_smoke(k):
    a = 0.8
    pk = ccl.nonlin_matter_power(COSMO, k, a)
    assert np.all(np.isfinite(pk))
    assert np.shape(pk) == np.shape(k)


@pytest.mark.parametrize('r', [
    1,
    1.0,
    [0.3, 0.5, 10],
    np.array([0.3, 0.5, 10])])
def test_sigmaR_smoke(r):
    a = 0.8
    sig = ccl.sigmaR(COSMO, r, a)
    assert np.all(np.isfinite(sig))
    assert np.shape(sig) == np.shape(r)


@pytest.mark.parametrize('r', [
    1,
    1.0,
    [0.3, 0.5, 10],
    np.array([0.3, 0.5, 10])])
def test_sigmaV_smoke(r):
    a = 0.8
    sig = ccl.sigmaV(COSMO, r, a)
    assert np.all(np.isfinite(sig))
    assert np.shape(sig) == np.shape(r)


def test_sigma8_consistent():
    assert np.allclose(ccl.sigma8(COSMO), COSMO['sigma8'])
    assert np.allclose(ccl.sigmaR(COSMO, 8 / COSMO['h'], 1), COSMO['sigma8'])


@pytest.mark.parametrize('A', [
    1,
    1.0,
    [0.3, 0.5, 1],
    np.array([0.3, 0.5, 1])])
def test_kNL(A):
    knl = ccl.kNL(COSMO, A)
    assert np.all(np.isfinite(knl))
    assert np.shape(knl) == np.shape(A)


@pytest.mark.parametrize('tf', [
    'boltzmann_class', 'boltzmann_camb', 'boltzmann_isitgr'])
def test_power_sigma8norm_norms_consistent(tf):
    # make a cosmo with A_s
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, A_s=2e-9, n_s=0.96,
        transfer_function=tf)
    sigma8 = ccl.sigma8(cosmo)

    # remake same but now give sigma8
    cosmo_s8 = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=sigma8, n_s=0.96,
        transfer_function=tf)

    # make sure they come out the same-ish
    assert np.allclose(sigma8, ccl.sigma8(cosmo_s8))

    # and that the power spectra look right
    a = 0.8
    gfac = (
        ccl.growth_factor(cosmo, a) / ccl.growth_factor(cosmo_s8, a))**2
    pk_rat = (
        ccl.linear_matter_power(cosmo, 1e-4, a) /
        ccl.linear_matter_power(cosmo_s8, 1e-4, a))
    assert np.allclose(pk_rat, gfac)


def test_input_lin_power_spectrum():
    # Setup
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7, n_s=0.965,
                          A_s=2e-9)
    a_arr = np.linspace(0.1, 1.0, 50)
    chi_from_ccl = ccl.background.comoving_radial_distance(cosmo, a_arr)
    hoh0_from_ccl = ccl.background.h_over_h0(cosmo, a_arr)
    growth_from_ccl = ccl.background.growth_factor_unnorm(cosmo, a_arr)
    fgrowth_from_ccl = ccl.background.growth_rate(cosmo, a_arr)
    k_arr = np.logspace(np.log10(2e-4), np.log10(1), 1000)
    pk_arr = np.empty(shape=(len(a_arr), len(k_arr)))
    for i, a in enumerate(a_arr):
        pk_arr[i] = ccl.power.linear_matter_power(cosmo, k_arr, a)

    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        background={'a': a_arr,
                    'chi': chi_from_ccl,
                    'h_over_h0': hoh0_from_ccl},
        growth={'a': a_arr,
                'growth_factor': growth_from_ccl,
                'growth_rate': fgrowth_from_ccl},
        pk_linear={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pk_arr})

    pk_CCL_input = ccl.power.linear_matter_power(cosmo_input, k_arr, 0.5)
    pk_CCL = ccl.power.linear_matter_power(cosmo, k_arr, 0.5)

    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Test again with negative power spectrum (so it's not logscaled)
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        background={'a': a_arr,
                    'chi': chi_from_ccl,
                    'h_over_h0': hoh0_from_ccl},
        growth={'a': a_arr,
                'growth_factor': growth_from_ccl,
                'growth_rate': fgrowth_from_ccl},
        pk_linear={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': -pk_arr})

    pk_CCL_input = -ccl.power.linear_matter_power(cosmo_input, k_arr, 0.5)

    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Via `linear_power`
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        background={'a': a_arr,
                    'chi': chi_from_ccl,
                    'h_over_h0': hoh0_from_ccl},
        growth={'a': a_arr,
                'growth_factor': growth_from_ccl,
                'growth_rate': fgrowth_from_ccl},
        pk_linear={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pk_arr,
                   'a:b': pk_arr})
    pk_CCL_input = ccl.power.linear_power(cosmo_input, k_arr, 0.5,
                                          p_of_k_a='a:b')
    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Shape mismatch of input arrays
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, A_s=2e-9,
            background={'a': a_arr,
                        'chi': chi_from_ccl,
                        'h_over_h0': hoh0_from_ccl[:-1]})


def test_input_linpower_raises():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7,
                          n_s=0.965, sigma8=0.8,
                          transfer_function='bbks')
    a_arr = np.linspace(0.1, 1.0, 50)
    k_arr = np.logspace(np.log10(2e-4), np.log10(1), 1000)
    pk_arr = np.array([ccl.power.linear_matter_power(cosmo, k_arr, a)
                       for a in a_arr])

    # Not a dictionary
    with pytest.raises(TypeError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear=np.pi)

    # a not increasing
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr[::-1], 'k': k_arr,
                       'delta_matter:delta_matter': pk_arr})

    # Dm x Dm not present
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'delta_matter;delta_matter': pk_arr})

    # Wrong shape
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'delta_matter:delta_matter': pk_arr,
                       'a:b': pk_arr[0]})

    # Wrong pk label
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'hello_there': pk_arr,
                       'a:b': pk_arr[0]})

    # Check new power spectrum is stored
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, sigma8=0.8,
        pk_linear={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pk_arr,
                   'a:b': pk_arr})
    assert 'a:b' in cosmo_input._pk_lin
    assert cosmo_input.has_linear_power


def test_input_nonlinear_model():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7, n_s=0.965,
                          A_s=2e-9, transfer_function='boltzmann_class')
    a_arr = np.linspace(0.1, 1.0, 50)
    k_arr = np.logspace(np.log10(2e-4), np.log10(1), 1000)
    pk_arr = np.empty(shape=(len(a_arr), len(k_arr)))
    for i, a in enumerate(a_arr):
        pk_arr[i] = ccl.power.nonlin_matter_power(cosmo, k_arr, a)

    pk_CCL = ccl.power.nonlin_matter_power(cosmo, k_arr, 0.5)

    # Test again passing only linear Pk, but letting HALOFIT do its thing
    kl_arr = np.logspace(-4, 1, 1000)
    pkl_arr = np.array([ccl.power.linear_matter_power(cosmo, kl_arr, a)
                        for a in a_arr])
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        pk_linear={'a': a_arr, 'k': kl_arr,
                   'delta_matter:delta_matter': pkl_arr},
        nonlinear_model='halofit')

    pk_CCL_input = ccl.power.nonlin_matter_power(cosmo_input, k_arr, 0.5)

    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Test extra power spectrum
    kl_arr = np.logspace(-4, 1, 1000)
    pkl_arr = np.array([ccl.power.linear_matter_power(cosmo, kl_arr, a)
                        for a in a_arr])
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        pk_linear={'a': a_arr, 'k': kl_arr,
                   'delta_matter:delta_matter': pkl_arr,
                   'a:b': pkl_arr},
        pk_nonlin={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': -pk_arr},
        nonlinear_model='halofit')

    pk_CCL_input = cosmo_input.get_nonlin_power('a:b')(k_arr, 0.5,
                                                       cosmo_input)
    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Via `nonlin_power`
    pk_CCL_input = ccl.power.nonlin_power(cosmo_input, k_arr, 0.5,
                                          p_of_k_a='a:b')
    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Use dictionary
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        pk_linear={'a': a_arr, 'k': kl_arr,
                   'delta_matter:delta_matter': pkl_arr,
                   'a:b': pkl_arr, 'c:d': pkl_arr},
        pk_nonlin={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': -pk_arr},
        nonlinear_model={'a:b': 'halofit',
                         'c:d': None})
    pk_CCL_input = ccl.power.nonlin_power(cosmo_input, k_arr, 0.5,
                                          p_of_k_a='a:b')
    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)
    assert 'c:d' not in cosmo_input._pk_nl


def test_input_nonlin_power_spectrum():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7, n_s=0.965,
                          A_s=2e-9, transfer_function='boltzmann_class')
    a_arr = np.linspace(0.1, 1.0, 50)
    chi_from_ccl = ccl.background.comoving_radial_distance(cosmo, a_arr)
    hoh0_from_ccl = ccl.background.h_over_h0(cosmo, a_arr)
    growth_from_ccl = ccl.background.growth_factor_unnorm(cosmo, a_arr)
    fgrowth_from_ccl = ccl.background.growth_rate(cosmo, a_arr)
    k_arr = np.logspace(np.log10(2e-4), np.log10(1), 1000)
    pk_arr = np.empty(shape=(len(a_arr), len(k_arr)))
    for i, a in enumerate(a_arr):
        pk_arr[i] = ccl.power.nonlin_matter_power(cosmo, k_arr, a)

    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        background={'a': a_arr,
                    'chi': chi_from_ccl,
                    'h_over_h0': hoh0_from_ccl},
        growth={'a': a_arr,
                'growth_factor': growth_from_ccl,
                'growth_rate': fgrowth_from_ccl},
        pk_nonlin={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pk_arr})

    pk_CCL_input = ccl.power.nonlin_matter_power(cosmo_input, k_arr, 0.5)
    pk_CCL = ccl.power.nonlin_matter_power(cosmo, k_arr, 0.5)

    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)

    # Test again with negative power spectrum (so it's not logscaled)
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, A_s=2e-9,
        background={'a': a_arr,
                    'chi': chi_from_ccl,
                    'h_over_h0': hoh0_from_ccl},
        growth={'a': a_arr,
                'growth_factor': growth_from_ccl,
                'growth_rate': fgrowth_from_ccl},
        pk_nonlin={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': -pk_arr})

    pk_CCL_input = -ccl.power.nonlin_matter_power(cosmo_input, k_arr, 0.5)

    assert np.allclose(pk_CCL_input, pk_CCL, atol=0., rtol=1e-5)


def test_input_nonlinear_model_raises():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7,
                          n_s=0.965, sigma8=0.8,
                          transfer_function='bbks')
    a_arr = np.linspace(0.1, 1.0, 50)
    k_arr = np.logspace(np.log10(2e-4), np.log10(1), 1000)
    pkl_arr = np.array([ccl.power.linear_matter_power(cosmo, k_arr, a)
                        for a in a_arr])

    # If no non-linear model provided, delta_matter:delta_matter
    # should be there.
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_nonlin={'a': a_arr, 'k': k_arr,
                       'a:b': pkl_arr})

    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'delta_matter:delta_matter': pkl_arr},
            nonlinear_model=np.pi)

    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            nonlinear_model='halofit')

    with pytest.raises(KeyError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'delta_matter:delta_matter': pkl_arr},
            nonlinear_model={'y:z': 'halofit'})

    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'delta_matter:delta_matter': pkl_arr},
            nonlinear_model={'delta_matter:delta_matter': None})

    with pytest.raises(KeyError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_linear={'a': a_arr, 'k': k_arr,
                       'delta_matter:delta_matter': pkl_arr},
            nonlinear_model={'delta_matter:delta_matter': 'unknown_model'})


def test_input_nonlin_raises():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7,
                          n_s=0.965, sigma8=0.8,
                          transfer_function='bbks')
    a_arr = np.linspace(0.1, 1.0, 50)
    k_arr = np.logspace(np.log10(2e-4), np.log10(1), 1000)
    pkl_arr = np.array([ccl.power.linear_matter_power(cosmo, k_arr, a)
                        for a in a_arr])
    pk_arr = np.array([ccl.power.nonlin_matter_power(cosmo, k_arr, a)
                       for a in a_arr])

    # Not a dictionary
    with pytest.raises(TypeError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_nonlin=np.pi)

    # k not present
    with pytest.raises(KeyError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_nonlin={'a': a_arr, 'kk': k_arr,
                       'delta_matter;delta_matter': pk_arr})

    # a not increasing
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_nonlin={'a': a_arr[::-1], 'k': k_arr,
                       'delta_matter:delta_matter': pk_arr})

    # delta_matter:delta_matter not present
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_nonlin={'a': a_arr, 'k': k_arr,
                       'delta_matter;delta_matter': pk_arr})

    # Wrong shape
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            pk_nonlin={'a': a_arr, 'k': k_arr,
                       'delta_matter:delta_matter': pk_arr,
                       'a:b': pk_arr[0]})

    # Linear Pk not set for halofit
    with pytest.raises(ValueError):
        ccl.CosmologyCalculator(
            Omega_c=0.27, Omega_b=0.05, h=0.7,
            n_s=0.965, sigma8=0.8,
            nonlinear_model='halofit')

    # Check new power spectrum is stored
    cosmo_input = ccl.CosmologyCalculator(
        Omega_c=0.27, Omega_b=0.05, h=0.7,
        n_s=0.965, sigma8=0.8,
        pk_linear={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pkl_arr,
                   'a:b': pkl_arr},
        pk_nonlin={'a': a_arr, 'k': k_arr,
                   'delta_matter:delta_matter': pk_arr},
        nonlinear_model='halofit')
    assert 'a:b' in cosmo_input._pk_nl
    assert cosmo_input.has_nonlin_power


def test_camb_de_model():
    """Check that the dark energy model for CAMB has been properly defined."""
    with pytest.raises(ValueError):
        cosmo = ccl.CosmologyVanillaLCDM(
            transfer_function='boltzmann_camb',
            extra_parameters={"camb": {"dark_energy_model": "pf"}})
        ccl.linear_matter_power(cosmo, 1, 1)

    """Check that w is not less than -1, if the chosen dark energy model for
    CAMB is fluid."""
    with pytest.raises(ValueError):
        cosmo = ccl.CosmologyVanillaLCDM(
            transfer_function='boltzmann_camb', w0=-1, wa=-1)
        ccl.linear_matter_power(cosmo, 1, 1)

    """Check that ppf is running smoothly."""
    cosmo = ccl.CosmologyVanillaLCDM(
        transfer_function='boltzmann_camb', w0=-1, wa=-1,
        extra_parameters={"camb": {"dark_energy_model": "ppf"}})
    assert np.isfinite(ccl.linear_matter_power(cosmo, 1, 1))



import numpy as np
import pytest

import pyccl as ccl

# NOTE: We now check up to kmax=23
# because CLASS v3 has a slight mismatch
# (6e-3) at higher k wavenumbers.
KMAX = 23
POWER_NU_TOL = 1.0E-3


@pytest.mark.parametrize('model', [0, 1, 2])
def test_power_nu(model):
    mnu = [[0.04, 0., 0.],
           [0.05, 0.01, 0.],
           [0.03, 0.02, 0.04]]
    w_0 = [-1.0, -0.9, -0.9]
    w_a = [0.0, 0.0, 0.1]

    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        A_s=2.1e-9,
        n_s=0.96,
        Neff=3.046, T_CMB=2.725,
        Omega_k=0,
        w0=w_0[model],
        wa=w_a[model],
        m_nu=mnu[model],
        mass_split='list',
        transfer_function='boltzmann_class')

    a = 1

    data_lin = np.loadtxt("./benchmarks/data/model%d_pk_nu.txt" % (model+1))
    k_lin = data_lin[:, 0] * cosmo['h']
    pk_lin = data_lin[:, 1] / (cosmo['h']**3)

    pk_lin_ccl = ccl.linear_matter_power(cosmo, k_lin, a)

    assert np.allclose(pk_lin_ccl[k_lin < KMAX],
                       pk_lin[k_lin < KMAX],
                       rtol=POWER_NU_TOL)

    data_nl = np.loadtxt("./benchmarks/data/model%d_pk_nl_nu.txt" % (model+1))
    k_nl = data_nl[:, 0] * cosmo['h']
    pk_nl = data_nl[:, 1] / (cosmo['h']**3)
    pk_nl_ccl = ccl.nonlin_matter_power(cosmo, k_nl, a)
    assert np.allclose(pk_nl_ccl, pk_nl, rtol=POWER_NU_TOL)



__all__ = ("MassFuncSheth99",)

import numpy as np

from . import MassFunc, get_delta_c


class MassFuncSheth99(MassFunc):
    """Implements the mass function of `Sheth & Tormen 1999
    <https://arxiv.org/abs/astro-ph/9901122>`_.
    This parametrization is only valid for 'fof' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
        use_delta_c_fit (:obj:`bool`): if ``True``, use the fit to the
            critical overdensity :math:`\\delta_c` by
            `Nakamura & Suto 1997
            <https://arxiv.org/abs/astro-ph/9612074>`_. Otherwise use
            :math:`\\delta_c = 1.68647`.
    """
    __repr_attrs__ = __eq_attrs__ = ("mass_def", "mass_def_strict",
                                     "use_delta_c_fit",)
    name = 'Sheth99'

    def __init__(self, *,
                 mass_def="fof",
                 mass_def_strict=True,
                 use_delta_c_fit=False):
        self.use_delta_c_fit = use_delta_c_fit
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta != "fof"

    def _setup(self):
        self.A = 0.21615998645
        self.p = 0.3
        self.a = 0.707

    def _get_fsigma(self, cosmo, sigM, a, lnM):
        if self.use_delta_c_fit:
            delta_c = get_delta_c(cosmo, a, 'NakamuraSuto97')
        else:
            delta_c = get_delta_c(cosmo, a, 'EdS')

        nu = delta_c / sigM
        return nu * self.A * (1. + (self.a * nu**2)**(-self.p)) * (
            np.exp(-self.a * nu**2/2.))



__all__ = ("HaloBiasBhattacharya11",)

from . import HaloBias, get_delta_c


class HaloBiasBhattacharya11(HaloBias):
    """ Implements halo bias as described in `Bhattacharya et al. 2011
    <https://arxiv.org/abs/1005.2239>`_. This parametrization is only
    valid for 'fof' masses.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
        mass_def_strict (:obj:`bool`): if ``False``, consistency of the mass
            definition will be ignored.
    """
    name = "Bhattacharya11"

    def __init__(self, *,
                 mass_def="fof",
                 mass_def_strict=True):
        super().__init__(mass_def=mass_def, mass_def_strict=mass_def_strict)

    def _check_mass_def_strict(self, mass_def):
        return mass_def.Delta != "fof"

    def _setup(self):
        self.a = 0.788
        self.az = 0.01
        self.p = 0.807
        self.q = 1.795
        self.dc = get_delta_c(None, None, kind='EdS')

    def _get_bsigma(self, cosmo, sigM, a):
        nu = self.dc / sigM
        a = self.a * a**self.az
        anu2 = a * nu**2
        return 1 + (anu2 - self.q + 2*self.p / (1 + anu2**self.p)) / self.dc



__all__ = ("deprecated", "deprecate_attr",)

import functools
import warnings

from .. import CCLDeprecationWarning


def deprecated(new_function=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used. If there is a replacement function,
    pass it as `new_function`.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            s = f"The function {func.__qualname__} is deprecated."
            if new_function:
                s += f" Use {new_function.__qualname__} instead."
            warnings.warn(s, CCLDeprecationWarning)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def deprecate_attr(getter=None, *, pairs=[]):
    """This decorator can be used to deprecate attributes,
    warning users about it and pointing them to the new attribute.

    Parameters
    ----------
    getter : slot wrapper ``__getattribute__``
        This is the getter method to be decorated.
    pairs : list of pairs
        List of renaming pairs ``('old', 'new')``.

    Example
    -------
    We have the legacy attribute ``old_name`` which we want to rename
    to ``new_name``. To achieve this we decorate the ``__getattribute__``
    method of the parent class in the main class body to retrieve the
    ``__getattr__`` method for the main class, like so:

    >>>  __getattr__ = deprecate_attr([('old_name', 'new_name')])(
             super.__getattribute__)

    Now, every time the attribute is called via its old name, the user will
    be warned about the renaming, and the attribute value will be returned.

    .. note:: Make sure that you bind ``__getattr__`` to the decorator,
              rather than ``__getattribute__``, because ``__getattr__``
              provides the fallback mechanism we want to use. Otherwise,
              an infinite recursion will initiate.

    """
    if getter is None:
        return functools.partial(deprecate_attr, pairs=pairs)

    rename = dict(pairs)

    @functools.wraps(getter)
    def wrapper(cls, name):
        if name in rename:
            new_name = rename[name]
            class_name = cls.__class__.__name__
            warnings.warn(
                f"Attribute {name} is deprecated in {class_name}. "
                f"Pass the new name {new_name}.", CCLDeprecationWarning)
            name = new_name

        return cls.__getattribute__(name)
    return wrapper



__all__ = ("Baryons",)

from abc import abstractmethod

from .. import CCLAutoRepr, CCLNamedClass


class Baryons(CCLAutoRepr, CCLNamedClass):
    """:class:`Baryons` objects are used to include the effects of
    baryons on the non-linear matter power spectrum. Their main ingredient
    is a method :meth:`include_baryonic_effects` that takes in a
    :class:`~pyccl.pk2d.Pk2D` and returns another :class:`~pyccl.pk2d.Pk2D`
    object that now accounts for baryonic effects (according to the model
    implemented in the corresponding :class:`Baryons` object).
    """

    @abstractmethod
    def _include_baryonic_effects(self, cosmo, pk):
        """Apply baryonic effects to a given power spectrum.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                Cosmological parameters.
            pk (:class:`~pyccl.pk2d.Pk2D`): power spectrum.

        Returns:
            :obj:`~pyccl.pk2d.Pk2D` object.
        """

    def include_baryonic_effects(self, cosmo, pk):
        """Apply baryonic effects to a given power spectrum.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`):
                Cosmological parameters.
            pk (:class:`~pyccl.pk2d.Pk2D`): power spectrum.

        Returns:
            :obj:`~pyccl.pk2d.Pk2D` or :obj:`None`.
        """
        return self._include_baryonic_effects(cosmo, pk)



import numpy as np
import pyccl as ccl

KNL_TOLERANCE = 1.0e-5


def test_kNL():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        sigma8=0.8,
        n_s=0.96,
        Neff=0,
        m_nu=0.0,
        w0=-1.,
        wa=0.,
        T_CMB=2.7,
        mass_split='normal',
        Omega_g=0,
        Omega_k=0,
        transfer_function='bbks',
        matter_power_spectrum='linear')

    data = np.loadtxt('./benchmarks/data/kNL.txt')
    a = data[:, 0]
    kNL = data[:, 1]
    kNL_ccl = ccl.kNL(cosmo, a)
    for i in range(len(a)):
        err = np.abs(kNL_ccl[i]/kNL[i] - 1)
        assert np.allclose(err, 0, rtol=0, atol=KNL_TOLERANCE)



__all__ = ("BaccoemuLinear",)

import numpy as np

from .. import Pk2D
from . import EmulatorPk


class BaccoemuLinear(EmulatorPk):
    """ Linear power spectrum emulator from baccoemu

    This is an emulator of the linear matter power spectrum as a function
    of 8 cosmological parameters (:math:`\\omega_c`, :math:`\\omega_b`,
    :math:`A_s` or :math:`\\sigma_{8,c}`, :math:`n_s`, :math:`h`,
    :math:`M_\\nu`, :math:`w_0`, :math:`w_a`) and the scale factor.

    See `Arico et al. 2021 <https://arxiv.org/abs/2104.14568>`_
    and https://bacco.dipc.org/emulator.html
    """
    def __init__(self):
        # avoid tensorflow warnings
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning)
            import baccoemu
            self.mpk = baccoemu.Matter_powerspectrum()
        self.a_min = self.mpk.emulator['linear']['bounds'][-1][0]
        self.a_max = self.mpk.emulator['linear']['bounds'][-1][1]
        self.k_min = self.mpk.emulator['linear']['k'][0]
        self.k_max = self.mpk.emulator['linear']['k'][-1]

    def __str__(self) -> str:
        return """baccoemu linear Pk module,
k_min,k_max = ({}, {}),
a_min,a_max = ({}, {})""".format(
            self.k_min, self.k_max, self.a_min, self.a_max)

    def _sigma8tot_2_sigma8cold(self, emupars, sigma8tot):
        """Use baccoemu to convert sigma8 total matter to sigma8 cdm+baryons
        """
        if hasattr(emupars['omega_cold'], '__len__'):
            _emupars = {}
            for pname in emupars:
                _emupars[pname] = emupars[pname][0]
        else:
            _emupars = emupars
        A_s_fid = 2.1e-9
        sigma8tot_fid = self.mpk.get_sigma8(cold=False, A_s=A_s_fid,
                                            **_emupars)
        A_s = (sigma8tot / sigma8tot_fid)**2 * A_s_fid
        return self.mpk.get_sigma8(cold=True, A_s=A_s, **_emupars)

    def _get_pk_at_a(self, cosmo, a):
        # First create the dictionary passed to baccoemu
        # if a is an array, make sure all the other parameters passed to the
        # emulator have the same len
        if hasattr(a, '__len__'):
            emupars = dict(
                omega_cold=np.full((len(a)),
                                   cosmo['Omega_c'] + cosmo['Omega_b']),
                omega_baryon=np.full((len(a)), cosmo['Omega_b']),
                ns=np.full((len(a)), cosmo['n_s']),
                hubble=np.full((len(a)), cosmo['h']),
                neutrino_mass=np.full((len(a)), np.sum(cosmo['m_nu'])),
                w0=np.full((len(a)), cosmo['w0']),
                wa=np.full((len(a)), cosmo['wa']),
                expfactor=a
            )
        else:
            emupars = dict(
                omega_cold=cosmo['Omega_c'] + cosmo['Omega_b'],
                omega_baryon=cosmo['Omega_b'],
                ns=cosmo['n_s'],
                hubble=cosmo['h'],
                neutrino_mass=np.sum(cosmo['m_nu']),
                w0=cosmo['w0'],
                wa=cosmo['wa'],
                expfactor=a
            )

        # if cosmo contains sigma8, we use it for baccoemu, otherwise we pass
        # A_s to the emulator
        if np.isnan(cosmo['A_s']):
            sigma8tot = cosmo['sigma8']
            sigma8cold = self._sigma8tot_2_sigma8cold(emupars, sigma8tot)
            if hasattr(a, '__len__'):
                emupars['sigma8_cold'] = np.full((len(a)), sigma8cold)
            else:
                emupars['sigma8_cold'] = sigma8cold
        else:
            if hasattr(a, '__len__'):
                emupars['A_s'] = np.full((len(a)), cosmo['A_s'])
            else:
                emupars['A_s'] = cosmo['A_s']

        h = cosmo['h']
        k_hubble, pk_hubble = self.mpk.get_linear_pk(cold=False, **emupars)
        return k_hubble * h, pk_hubble / h**3

    def _get_pk2d(self, cosmo):
        a = cosmo.get_pk_spline_a()
        a_for_baccoemu = a[a >= self.a_min]
        a_extrapolated = a[a < self.a_min]
        # we directly use the emulator for the expansion factors within its
        # range
        k, pk = self.get_pk_at_a(cosmo, a_for_baccoemu)
        # for the expansion factors requested by ccl but outside the emulator
        # range, we extrapolate from the earliest pk available with linear
        # growth factors.
        # NOTE: ccl computes scale independent growth factors, this is not
        # correct with massive neutrinos
        growth_factors = cosmo.growth_factor(a_extrapolated)
        ref_growth_factor = cosmo.growth_factor(a_for_baccoemu[0])
        pk_extrapolated = (np.repeat([pk[0]], [len(a_extrapolated)], axis=0).T
                           * (growth_factors / ref_growth_factor)**2).T
        # now we combine the extrapolated and direct spectra
        pk_final = np.concatenate([pk_extrapolated, pk])
        return Pk2D(a_arr=a, lk_arr=np.log(k), pk_arr=np.log(pk_final),
                    is_logp=True, extrap_order_lok=1, extrap_order_hik=2)



__all__ = ("angular_cl_cov_cNG", "sigma2_B_disc", "sigma2_B_from_mask",
           "angular_cl_cov_SSC",)

import numpy as np

from . import DEFAULT_POWER_SPECTRUM, check, lib
from .pyutils import _check_array_params, integ_types


def angular_cl_cov_cNG(cosmo, tracer1, tracer2, *, ell, t_of_kk_a,
                       tracer3=None, tracer4=None, ell2=None,
                       fsky=1., integration_method='qag_quad'):
    """Calculate the connected non-Gaussian covariance for a pair of
    angular power spectra :math:`C_{\\ell_1}^{ab}` and :math:`C_{\\ell_2}^{cd}`,
    between two pairs of tracers (:math:`(a,b)` and :math:`(c,d)`).

    Specifically, it computes:

    .. math::
        {\\rm Cov}_{\\rm cNG}(\\ell_1,\\ell_2)=\\frac{1}{4\\pi f_{\\rm sky}}
        \\int \\frac{d\\chi}{\\chi^6}
        \\tilde{\\Delta}^a_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^b_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^c_{\\ell_2}(\\chi)
        \\tilde{\\Delta}^d_{\\ell_2}(\\chi)\\,
        \\bar{T}_{abcd}\\left[\\frac{\\ell_1+1/2}{\\chi},
                              \\frac{\\ell_2+1/2}{\\chi}, a(\\chi)\\right]

    where :math:`\\Delta^x_\\ell(\\chi)` is the transfer function for tracer
    :math:`x` (see Eq. 39 in the CCL note), and
    :math:`\\bar{T}_{abcd}(k_1,k_2,a)` is the isotropized connected
    trispectrum of the four tracers (see the documentation of the
    :class:`~pyccl.tk3d.Tk3D` class for details).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        tracer1 (:class:`~pyccl.tracers.Tracer`): a Tracer object.
        tracer2 (:class:`~pyccl.tracers.Tracer`): a second Tracer object.
        ell (:obj:`float` or `array`): Angular wavenumber(s) at which to evaluate
            the first dimension of the angular power spectrum covariance.
        t_of_kk_a (:class:`~pyccl.tk3d.Tk3D`): 3D connected
            trispectrum.
        tracer3 (:class:`~pyccl.tracers.Tracer`): a Tracer object.
            If ``None``, ``tracer1`` will be used instead.
        tracer4 (:class:`~pyccl.tracers.Tracer`): a Tracer object.
            If ``None``, ``tracer2`` will be used instead.
        ell2 (:obj:`float` or `array`): Angular wavenumber(s) at which to evaluate
            the second dimension of the angular power spectrum covariance. If
            ``None``, ``ell`` will be used instead.
        fsky (:obj:`float`): sky fraction.
        integration_method (:obj:`str`) : integration method to be used
            for the Limber integrals. Possibilities: ``'qag_quad'`` (GSL's
            `qag` method backed up by `quad` when it fails) and ``'spline'``
            (the integrand is splined and then integrated analytically).

    Returns:
        (:obj:`float` or `array`): 2D array containing the connected non-Gaussian \
            Angular power spectrum covariance \
            :math:`{\\rm Cov}_{\\rm cNG}(\\ell_1,\\ell_2)`, for the \
            four input tracers, as a function of :math:`\\ell_1` and \
            :math:`\\ell_2`. The ordering is such that \
            ``out[i2, i1] = Cov(ell2[i2], ell[i1])``.
    """ # noqa
    if integration_method not in integ_types:
        raise ValueError(f"Unknown integration method {integration_method}.")

    # we need the distances for the integrals
    cosmo.compute_distances()

    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    tsp = t_of_kk_a.tsp

    # Create tracer colections
    status = 0
    clt1, status = lib.cl_tracer_collection_t_new(status)
    for t in tracer1._trc:
        status = lib.add_cl_tracer_to_collection(clt1, t, status)
    clt2, status = lib.cl_tracer_collection_t_new(status)
    for t in tracer2._trc:
        status = lib.add_cl_tracer_to_collection(clt2, t, status)
    if tracer3 is None:
        clt3 = clt1
    else:
        clt3, status = lib.cl_tracer_collection_t_new(status)
        for t in tracer3._trc:
            status = lib.add_cl_tracer_to_collection(clt3, t, status)
    if tracer4 is None:
        clt4 = clt2
    else:
        clt4, status = lib.cl_tracer_collection_t_new(status)
        for t in tracer4._trc:
            status = lib.add_cl_tracer_to_collection(clt4, t, status)

    ell1_use = np.atleast_1d(ell)
    if ell2 is None:
        ell2 = ell
    ell2_use = np.atleast_1d(ell2)

    cov, status = lib.angular_cov_vec(
        cosmo, clt1, clt2, clt3, clt4, tsp,
        ell1_use, ell2_use, integ_types[integration_method],
        6, 1./(4*np.pi*fsky), ell1_use.size*ell2_use.size, status)

    cov = cov.reshape([ell2_use.size, ell1_use.size])
    if np.ndim(ell2) == 0:
        cov = np.squeeze(cov, axis=0)
    if np.ndim(ell) == 0:
        cov = np.squeeze(cov, axis=-1)

    # Free up tracer collections
    lib.cl_tracer_collection_t_free(clt1)
    lib.cl_tracer_collection_t_free(clt2)
    if tracer3 is not None:
        lib.cl_tracer_collection_t_free(clt3)
    if tracer4 is not None:
        lib.cl_tracer_collection_t_free(clt4)

    check(status, cosmo=cosmo_in)
    return cov


def sigma2_B_disc(cosmo, a_arr=None, *, fsky=1.,
                  p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """Returns the variance of the projected linear density field
    over a circular disc covering a sky fraction `fsky` as a function
    of scale factor. This is given by

    .. math::
        \\sigma^2_B(z) = \\int_0^\\infty \\frac{k\\,dk}{2\\pi}
            P_L(k,z)\\,\\left[\\frac{2J_1(k R(z))}{k R(z)}\\right]^2,

    where :math:`R(z)` is the corresponding radial aperture as a
    function of redshift. This quantity can be used to compute the
    super-sample covariance (see :func:`angular_cl_cov_SSC`).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        a_arr (:obj:`float`, `array` or :obj:`None`): an array of scale factor
            values at which to evaluate the projected variance. If
            ``None``, a default sampling will be used.
        fsky (:obj:`float`): sky fraction.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D` or :obj:`str`): Linear
            power spectrum to use. Defaults to the
            internal linear power spectrum from ``cosmo``.

    Returns:
        Tuple containing

        - a_arr (`array`): an array of scale factor values at which the
          projected variance has been evaluated. Only returned if ``a_arr``
          is ``None`` on input.
        - sigma2_B (:obj:`float` or `array`): projected variance.
    """
    full_output = a_arr is None

    if full_output:
        a_arr = cosmo.get_pk_spline_a()
    else:
        ndim = np.ndim(a_arr)
        a_arr = np.atleast_1d(a_arr)

    chi_arr = cosmo.comoving_radial_distance(a_arr)
    R_arr = chi_arr * np.arccos(1-2*fsky)
    psp = cosmo.parse_pk2d(p_of_k_a, is_linear=True)

    status = 0
    s2B_arr, status = lib.sigma2b_vec(cosmo.cosmo, a_arr, R_arr, psp,
                                      len(a_arr), status)
    check(status, cosmo=cosmo)

    if full_output:
        return a_arr, s2B_arr
    if ndim == 0:
        return s2B_arr[0]
    return s2B_arr


def sigma2_B_from_mask(cosmo, a_arr=None, *, mask_wl=None,
                       p_of_k_a=DEFAULT_POWER_SPECTRUM):
    """ Returns the variance of the projected linear density field, given the
    angular power spectrum of the footprint mask and scale factor. This is
    given by

    .. math::
        \\sigma^2_B(z) = \\frac{1}{\\chi^2(z)}\\sum_\\ell
            P_L\\left(\\frac{\\ell+\\frac{1}{2}}{\\chi(z)},z\\right)\\,
            \\sum_{m=-\\ell}^\\ell W^A_{\\ell m} {W^B}^*_{\\ell m},

    where :math:`W^A_{\\ell m}` and :math:`W^B_{\\ell m}` are the spherical
    harmonics decomposition of the footprint masks of fields `A` and `B`,
    normalized by their area. This quantity can be used to compute the
    super-sample covariance (see :func:`angular_cl_cov_SSC`).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): a Cosmology object.
        a_arr (:obj:`float`, `array` or :obj:`None`): an array of scale factor
            values at which to evaluate the projected variance.
        mask_wl (`array`): Array with the angular power spectrum of the
            masks. The power spectrum should be given at integer multipoles,
            starting at :math:`\\ell=0`. The power spectrum is normalized as
            :math:`{\\tt mask\\_wl}=\\sum_m W^A_{\\ell m} {W^B}^*_{\\ell m}`.
            It is the responsibility of the user to the provide the mask power
            out to sufficiently high ell for their required precision.
        p_of_k_a (:class:`~pyccl.pk2d.Pk2D` or :obj:`str`): Linear
            power spectrum to use. Defaults to the
            internal linear power spectrum from `cosmo`.

    Returns:
        Tuple containing

        - a_arr (`array`): an array of scale factor values at which the
          projected variance has been evaluated. Only returned if ``a_arr`` is
          ``None`` on input.
        - sigma2_B (:obj:`float` or `array`): projected variance.
    """
    full_output = a_arr is None

    if full_output:
        a_arr = cosmo.get_pk_spline_a()
    else:
        ndim = np.ndim(a_arr)
        a_arr = np.atleast_1d(a_arr)

    if p_of_k_a is DEFAULT_POWER_SPECTRUM:
        cosmo.compute_linear_power()
        p_of_k_a = cosmo.get_linear_power()

    ell = np.arange(mask_wl.size)

    sigma2_B = np.zeros(a_arr.size)
    for i in range(sigma2_B.size):
        if 1-a_arr[i] < 1e-6:
            # For a=1, the integral becomes independent of the footprint in
            # the flat-sky approximation. So we are just using the method
            # for the disc geometry here
            sigma2_B[i] = sigma2_B_disc(cosmo=cosmo, a_arr=a_arr[i],
                                        p_of_k_a=p_of_k_a)
        else:
            chi = cosmo.comoving_angular_distance(a=a_arr)
            k = (ell+0.5)/chi[i]
            pk = p_of_k_a(k, a_arr[i], cosmo)
            # See eq. E.10 of 2007.01844
            sigma2_B[i] = np.sum(pk * mask_wl)/chi[i]**2

    if full_output:
        return a_arr, sigma2_B
    if ndim == 0:
        return sigma2_B[0]
    return sigma2_B


def angular_cl_cov_SSC(cosmo, tracer1, tracer2, *, ell, t_of_kk_a,
                       tracer3=None, tracer4=None, ell2=None,
                       sigma2_B=None, fsky=1.,
                       integration_method='qag_quad'):
    """Calculate the super-sample contribution to the connected
    non-Gaussian covariance for a pair of power spectra
    :math:`C_{\\ell_1}^{ab}` and :math:`C_{\\ell_2}^{cd}`,
    between two pairs of tracers (:math:`(a,b)` and :math:`(c,d)`).

    Specifically, it computes:

    .. math::
        {\\rm Cov}_{\\rm SSC}(\\ell_1,\\ell_2)=
        \\int \\frac{d\\chi}{\\chi^4}
        \\tilde{\\Delta}^a_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^b_{\\ell_1}(\\chi)
        \\tilde{\\Delta}^c_{\\ell_2}(\\chi)
        \\tilde{\\Delta}^d_{\\ell_2}(\\chi)\\,
        \\bar{T}_{abcd}\\left[\\frac{\\ell_1+1/2}{\\chi},
                              \\frac{\\ell_2+1/2}{\\chi}, a(\\chi)\\right]

    where :math:`\\Delta^x_\\ell(\\chi)` is the transfer function for tracer
    :math:`x` (see Eq. 39 in the CCL note), and
    :math:`\\bar{T}_{abcd}(k_1,k_2,a)` is the isotropized connected
    trispectrum of the four tracers (see the documentation of the
    :class:`~pyccl.tk3d.Tk3D` class for details).

    Args:
        cosmo (:class:`~pyccl.cosmology.Cosmology`): A Cosmology object.
        tracer1 (:class:`~pyccl.tracers.Tracer`): a Tracer object.
        tracer2 (:class:`~pyccl.tracers.Tracer`): a second Tracer object.
        ell (:obj:`float` or `array`): Angular wavenumber(s) at which to evaluate
            the first dimension of the angular power spectrum covariance.
        t_of_kk_a (:class:`~pyccl.tk3d.Tk3D`): 3D connected
            trispectrum.
        tracer3 (:class:`~pyccl.tracers.Tracer`): a Tracer object.
            If ``None``, ``tracer1`` will be used instead.
        tracer4 (:class:`~pyccl.tracers.Tracer`): a Tracer object.
            If ``None``, ``tracer2`` will be used instead.
        ell2 (:obj:`float` or `array`): Angular wavenumber(s) at which to evaluate
            the second dimension of the angular power spectrum covariance. If
            ``None``, ``ell`` will be used instead.
        sigma2_B (:obj:`tuple` or :obj:`None`): A tuple of arrays
            (a, sigma2_B(a)) containing the variance of the projected matter
            overdensity over the footprint as a function of the scale factor.
            If ``None``, a compact circular footprint will be assumed covering
            a sky fraction ``fsky``.
        fsky (:obj:`float`): sky fraction.
        integration_method (:obj:`str`) : integration method to be used
            for the Limber integrals. Possibilities: ``'qag_quad'`` (GSL's
            `qag` method backed up by `quad` when it fails) and ``'spline'``
            (the integrand is splined and then integrated analytically).

    Returns:
        (:obj:`float` or `array`): 2D array containing the super-sample \
            Angular power spectrum covariance \
            :math:`{\\rm Cov}_{\\rm SSC}(\\ell_1,\\ell_2)`, for the \
            four input tracers, as a function of :math:`\\ell_1` and \
            :math:`\\ell_2`. The ordering is such that \
            ``out[i2, i1] = Cov(ell2[i2], ell[i1])``.
    """ # noqa
    if integration_method not in integ_types:
        raise ValueError(f"Unknown integration method {integration_method}.")

    # we need the distances for the integrals
    cosmo.compute_distances()

    # Access ccl_cosmology object
    cosmo_in = cosmo
    cosmo = cosmo.cosmo

    tsp = t_of_kk_a.tsp

    # Create tracer colections
    status = 0
    clt1, status = lib.cl_tracer_collection_t_new(status)
    for t in tracer1._trc:
        status = lib.add_cl_tracer_to_collection(clt1, t, status)
    clt2, status = lib.cl_tracer_collection_t_new(status)
    for t in tracer2._trc:
        status = lib.add_cl_tracer_to_collection(clt2, t, status)
    if tracer3 is None:
        clt3 = clt1
    else:
        clt3, status = lib.cl_tracer_collection_t_new(status)
        for t in tracer3._trc:
            status = lib.add_cl_tracer_to_collection(clt3, t, status)
    if tracer4 is None:
        clt4 = clt2
    else:
        clt4, status = lib.cl_tracer_collection_t_new(status)
        for t in tracer4._trc:
            status = lib.add_cl_tracer_to_collection(clt4, t, status)

    ell1_use = np.atleast_1d(ell)
    if ell2 is None:
        ell2 = ell
    ell2_use = np.atleast_1d(ell2)

    if sigma2_B is None:
        a_arr, s2b_arr = sigma2_B_disc(cosmo_in, fsky=fsky)
    else:
        a_arr, s2b_arr = _check_array_params(sigma2_B, 'sigma2_B')
    cov, status = lib.angular_cov_ssc_vec(
        cosmo, clt1, clt2, clt3, clt4, tsp, a_arr, s2b_arr,
        ell1_use, ell2_use, integ_types[integration_method],
        4, 1., ell1_use.size*ell2_use.size, status)

    cov = cov.reshape([ell2_use.size, ell1_use.size])
    if np.ndim(ell2) == 0:
        cov = np.squeeze(cov, axis=0)
    if np.ndim(ell) == 0:
        cov = np.squeeze(cov, axis=-1)

    # Free up tracer collections
    lib.cl_tracer_collection_t_free(clt1)
    lib.cl_tracer_collection_t_free(clt2)
    if tracer3 is not None:
        lib.cl_tracer_collection_t_free(clt3)
    if tracer4 is not None:
        lib.cl_tracer_collection_t_free(clt4)

    check(status, cosmo=cosmo_in)
    return cov



import os
import numpy as np
import pyccl as ccl

cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, Omega_g=0, Omega_k=0,
                      h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                      w0=-1, wa=0, T_CMB=2.7255,
                      transfer_function='eisenstein_hu')
dirdat = os.path.join(os.path.dirname(__file__), 'data')
Ms, Rs_200m, Rs_500c, Ms_500c = np.loadtxt(os.path.join(dirdat,
                                                        'mdef_bm.txt'),
                                           unpack=True)
dc = np.loadtxt(os.path.join(dirdat, 'conc_bm.txt'), unpack=True)
Ms = dc[0]
cs_200m_d = dc[1]
cs_200c_d = dc[2]
cs_200m_b = dc[3]
cs_200c_b = dc[4]
cs_vir_k = dc[5]
cs_vir_b = dc[6]
cs_200c_p = dc[7]
cs_200c_di = dc[8]

hmd_vir = ccl.halos.MassDefVir
hmd_200m = ccl.halos.MassDef200m
hmd_200c = ccl.halos.MassDef200c
hmd_500c = ccl.halos.MassDef500c


def test_mdef_get_radius():
    Rs_200m_h = hmd_200m.get_radius(cosmo, Ms, 1.)
    Rs_500c_h = hmd_500c.get_radius(cosmo, Ms, 1.)
    assert np.all(np.fabs(Rs_200m_h/Rs_200m-1) < 1E-6)
    assert np.all(np.fabs(Rs_500c_h/Rs_500c-1) < 1E-6)


def test_mdef_get_mass():
    Ms_h = hmd_200m.get_mass(cosmo, Rs_200m, 1.)
    assert np.all(np.fabs(Ms_h/Ms-1) < 1E-6)
    Ms_h = hmd_500c.get_mass(cosmo, Rs_500c, 1.)
    assert np.all(np.fabs(Ms_h/Ms-1) < 1E-6)


def get_cm(name, mass_def):
    return ccl.halos.Concentration.create_instance(name, mass_def=mass_def)


def test_mdef_concentration():
    cs_200m_dh = get_cm("Duffy08", "200m")(cosmo, Ms, 1)
    cs_200m_bh = get_cm("Bhattacharya13", "200m")(cosmo, Ms, 1)
    assert np.allclose(cs_200m_dh, cs_200m_d, atol=0, rtol=1e-6)
    assert np.allclose(cs_200m_bh, cs_200m_b, atol=0, rtol=3e-3)

    cs_200c_dh = get_cm("Duffy08", "200c")(cosmo, Ms, 1)
    cs_200c_bh = get_cm("Bhattacharya13", "200c")(cosmo, Ms, 1)
    assert np.allclose(cs_200c_dh, cs_200c_d, atol=0, rtol=1e-6)
    assert np.allclose(cs_200c_bh, cs_200c_b, atol=0, rtol=3e-3)

    cs_vir_kh = get_cm("Klypin11", "vir")(cosmo, Ms, 1)
    cs_vir_bh = get_cm("Bhattacharya13", "vir")(cosmo, Ms, 1)
    assert np.allclose(cs_vir_kh, cs_vir_k, atol=0, rtol=1e-6)
    assert np.allclose(cs_vir_bh, cs_vir_b, atol=0, rtol=3e-3)

    cs_200c_ph = get_cm("Prada12", "200c")(cosmo, Ms, 1)
    cs_200c_dih = get_cm("Diemer15", "200c")(cosmo, Ms, 1)
    assert np.allclose(cs_200c_ph, cs_200c_p, atol=0, rtol=1e-3)
    assert np.allclose(cs_200c_dih, cs_200c_di, atol=0, rtol=1e-2)


def test_mdef_translate_mass():
    translator = ccl.halos.mass_translator(mass_in=hmd_200m, mass_out=hmd_500c,
                                           concentration="Duffy08")
    Ms_500c_h = translator(cosmo, Ms, 1.)
    assert np.all(np.fabs(Ms_500c_h/Ms_500c-1) < 1E-6)
